Index: test/Verifier/gc_relocate_addrspace.ll
===================================================================
--- /dev/null
+++ test/Verifier/gc_relocate_addrspace.ll
@@ -0,0 +1,23 @@
+; RUN: not llvm-as -disable-output <%s 2>&1 | FileCheck %s
+; This is to verify that gc_relocate must return a pointer with the same
+; address space with the relocated value.
+
+; CHECK:       gc.relocate: relocating a pointer shouldn't change its address space
+; CHECK-NEXT:  %obj.relocated = call coldcc i8* @llvm.experimental.gc.relocate.p0i8(i32 %safepoint_token, i32 7, i32 7) ;
+
+declare void @foo()
+
+; Function Attrs: nounwind
+declare i32 @llvm.experimental.gc.statepoint.p0f_isVoidf(i64, i32, void ()*, i32, i32, ...) #0
+
+define void @test1(i64 addrspace(1)* %obj) gc "statepoint-example" {
+entry:
+  %safepoint_token = call i32 (i64, i32, void ()*, i32, i32, ...) @llvm.experimental.gc.statepoint.p0f_isVoidf(i64 0, i32 0, void ()* @foo, i32 0, i32 0, i32 0, i32 0, i64 addrspace(1)* %obj)
+  %obj.relocated = call coldcc i8* @llvm.experimental.gc.relocate.p0i8(i32 %safepoint_token, i32 7, i32 7) ; (%obj, %obj)
+  ret void
+}
+
+; Function Attrs: nounwind
+declare i8* @llvm.experimental.gc.relocate.p0i8(i32, i32, i32) #0
+
+attributes #0 = { nounwind }
Index: test/Verifier/gc_relocate_operand.ll
===================================================================
--- /dev/null
+++ test/Verifier/gc_relocate_operand.ll
@@ -0,0 +1,21 @@
+; RUN: not llvm-as -disable-output <%s 2>&1 | FileCheck %s
+; This is to verify that the relocated value by gc_relocate must be a pointer type.
+
+; CHECK: gc.relocate: relocated value must be a gc pointer
+
+declare void @foo()
+
+declare i32 @llvm.experimental.gc.statepoint.p0f_isVoidf(i64, i32, void ()*, i32, i32, ...)
+
+define void @test1(i64 %obj) gc "statepoint-example" {
+entry:
+  %safepoint_token = call i32 (i64, i32, void ()*, i32, i32, ...) @llvm.experimental.gc.statepoint.p0f_isVoidf(i64 0, i32 0, void ()* @foo, i32 0, i32 0, i32 0, i32 0, i64 %obj)
+  %obj.relocated = call coldcc i8 addrspace(1)* @llvm.experimental.gc.relocate.p1i8(i32 %safepoint_token, i32 7, i32 7) ; (%obj, %obj)
+  ret void
+}
+
+; Function Attrs: nounwind
+declare i8 addrspace(1)* @llvm.experimental.gc.relocate.p1i8(i32, i32, i32) #0
+
+attributes #0 = { nounwind }
+
Index: test/Verifier/gc_relocate_return.ll
===================================================================
--- /dev/null
+++ test/Verifier/gc_relocate_return.ll
@@ -0,0 +1,22 @@
+; RUN: not llvm-as -disable-output <%s 2>&1 | FileCheck %s
+; This is to verify that gc_relocate must return a pointer type, which is defined
+; in intrinsics.td.
+
+; CHECK: Intrinsic has incorrect return type!
+
+declare void @foo()
+
+declare i32 @llvm.experimental.gc.statepoint.p0f_isVoidf(i64, i32, void ()*, i32, i32, ...)
+
+define void @test1(<2 x i32 addrspace(1)*> addrspace(1)* %obj) gc "statepoint-example" {
+entry:
+  %safepoint_token = call i32 (i64, i32, void ()*, i32, i32, ...) @llvm.experimental.gc.statepoint.p0f_isVoidf(i64 0, i32 0, void ()* @foo, i32 0, i32 0, i32 0, i32 0, <2 x i32 addrspace(1)*> addrspace(1)* %obj)
+  %obj.relocated = call coldcc i8 @llvm.experimental.gc.relocate.i8(i32 %safepoint_token, i32 7, i32 7) ; (%obj, %obj)
+  ret void
+}
+
+; Function Attrs: nounwind
+declare i8 @llvm.experimental.gc.relocate.i8(i32, i32, i32) #0
+
+attributes #0 = { nounwind }
+
Index: lib/IR/Verifier.cpp
===================================================================
--- lib/IR/Verifier.cpp
+++ lib/IR/Verifier.cpp
@@ -3440,8 +3440,18 @@
            "'gc parameters' section of the statepoint call",
            &CI);
 
-    // gc_relocate does not need to be the same type as the relocated pointer.
-    // It can casted to the correct type later if it's desired
+    // Relocated value must be a pointer type, but gc_relocate does not need to return the
+    // same pointer type as the relocated pointer. It can be casted to the correct type later
+    // if it's desired. However, they must have the same address space.
+    GCRelocateOperands Operands(&CI);
+    Assert(Operands.getDerivedPtr()->getType()->isPointerTy(),
+           "gc.relocate: relocated value must be a gc pointer", &CI);
+
+    // gc_relocate return type must be a pointer type, and is verified earlier in
+    // VerifyIntrinsicType().
+    Assert(cast<PointerType>(CI.getType())->getAddressSpace() ==
+           cast<PointerType>(Operands.getDerivedPtr()->getType())->getAddressSpace(),
+           "gc.relocate: relocating a pointer shouldn't change its address space", &CI);
     break;
   }
   };